Attribute VB_Name = "mLens"
Option Explicit

Private Type XYCOORD
    x As Long
    y As Long
End Type

Private Type SAFEARRAYBOUND
    cElements As Long
    lLbound   As Long
End Type

Private Type SAFEARRAY2D
    cDims      As Integer
    fFeatures  As Integer
    cbElements As Long
    cLocks     As Long
    pvData     As Long
    Bounds(1)  As SAFEARRAYBOUND
End Type

Private Declare Function VarPtrArray Lib "msvbvm50" Alias "VarPtr" (Ptr() As Any) As Long
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (lpDst As Any, lpSrc As Any, ByVal ByteLength As Long)



'-- Lens offsets
Private LookUp() As XYCOORD

'-- LookUp table
Public Sub DIB32_Lens_BuildLookUpTable(ByVal Radius As Long, ByVal Factor As Long, Optional ByVal Convex As Boolean = -1)
  
  Dim i As Long, iL As Long
  Dim j As Long, jL As Long
  
  Dim cf As Single
  Dim rL As Long, sL As Single, zL As Long
  Dim rLpow2 As Long, sLpow2 As Long
        
    rL = Radius
    rLpow2 = rL * rL + 1

    cf = rL * (150 - Abs(Factor)) / 150
    sL = Sqr(rL * rL - cf * cf)
    rLpow2 = rL * rL
    sLpow2 = sL * sL
   
    ReDim LookUp(-rL To rL, -rL To rL)
            
    For j = -rL To rL
        For i = -rL To rL
            If (i * i + j * j < sLpow2) Then
                zL = Sqr(rLpow2 - i * i - j * j)
                If (Convex) Then
                    iL = (i * cf) \ zL
                    jL = (j * cf) \ zL
                  Else
                    iL = i - (i * cf) \ zL
                    jL = j - (j * cf) \ zL
                End If
                LookUp(i, j).x = iL
                LookUp(i, j).y = jL
              Else
                LookUp(i, j).x = i
                LookUp(i, j).y = j
            End If
        Next i
    Next j
End Sub

'-- Lens processing
Public Sub DIB32_Lens(DIB32Src As cDIB32, DIB32Dst As cDIB32)

  Dim sBits() As Long
  Dim dBits() As Long
  Dim stSA    As SAFEARRAY2D
  Dim dtSA    As SAFEARRAY2D
  
  Dim i As Long
  Dim j As Long
  
  Dim rL As Long
  
    rL = UBound(LookUp)

    If (DIB32Src.hDIB <> 0 And DIB32Dst.hDIB <> 0 And rL > 0) Then
    
        pvBuildSA stSA, DIB32Src
        CopyMemory ByVal VarPtrArray(sBits()), VarPtr(stSA), 4
        pvBuildSA dtSA, DIB32Dst
        CopyMemory ByVal VarPtrArray(dBits()), VarPtr(dtSA), 4

        For j = -rL To rL
            For i = -rL To rL
                dBits(i + rL, j + rL) = sBits(LookUp(i, j).x + rL, LookUp(i, j).y + rL)
            Next i
        Next j
        CopyMemory ByVal VarPtrArray(dBits), 0&, 4
        CopyMemory ByVal VarPtrArray(sBits), 0&, 4
    End If
End Sub

'==================================================================================================

Private Sub pvBuildSA(tSA As SAFEARRAY2D, DIB As cDIB32)
    With tSA
        .cbElements = IIf(App.LogMode = 1, 1, 4)
        .cDims = 2
        .Bounds(0).lLbound = 0
        .Bounds(0).cElements = DIB.Height
        .Bounds(1).lLbound = 0
        .Bounds(1).cElements = DIB.Width
        .pvData = DIB.DIBitsPtr
    End With
End Sub

